from __future__ import annotations
from typing import TYPE_CHECKING, List, Tuple, Optional
if TYPE_CHECKING:
    from syrinx.node import ContentNode
    from datetime import datetime
    ListOfUrls = List[Tuple[str, Optional[datetime]]]


def collect_urls(node: ContentNode) -> ListOfUrls:
    urls = []
    if node.includeInSitemap and node.address:
        urls.append((node.address, node.lastModified))

    for leaf in node.leaves:
        urls += collect_urls(leaf)

    for branch in node.branches:
        urls += collect_urls(branch)

    return urls


def generate_sitemap(urls: ListOfUrls) -> str:
    """Sitemap content string from list of url, datetime tuples
    """
    s = '<?xml version="1.0" encoding="UTF-8"?>\n'
    s += '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">\n'
    for (url, dt) in urls:
        lm = f'<lastmod>{dt.date().isoformat()}</lastmod>' if dt else ''
        s += f'    <url><loc>{url}</loc>{lm}</url>\n'
    s += '</urlset>'
    return s




from argparse import ArgumentParser, SUPPRESS
from syrinx.run import run_pipeline
from syrinx.scaffold import generate_scaffold
from syrinx.server.dev_server import DevServer


def get_args():
    """Use ArgumentParser to interpret CLI arguments and return an object with parsed settings.


    ```
    syrinx --help
    syrinx build
    syrinx build --dir meta
    syrinx serve --dir meta
    syrinx new --dir my_site
    ```
    """

    ## Shared arguments
    base_parser = ArgumentParser(add_help=False)
    base_parser.add_argument('-d', '--dir', type=str, default='.', help='Location of root directory to build from')
    base_parser.add_argument('-c', '--clean', default=SUPPRESS, action='store_true',
                        help='Remove existing dynamic content files')
    base_parser.add_argument('-e', '--environment', default=SUPPRESS, 
                        help='Define build environment for customization, e.g. "production"')
    base_parser.add_argument('--leaf-pages', default=SUPPRESS, action='store_true',
                        help='Build pages for "leaf" (non-index) content nodes')
    base_parser.add_argument('-v', '--verbose', default=SUPPRESS, action='store_true', 
                        help='Print log messages during build')

    ## Top level parser object
    root_parser = ArgumentParser()
    sub_parsers = root_parser.add_subparsers(
        title='Subcommands ("syrinx . <command> --help" for help for the specific command)')

    ## This adds the "build" sub-command
    build_parser = sub_parsers.add_parser('build', help='Generate html docs', parents=[base_parser])
    build_parser.set_defaults(command='build')

    ## This adds the "serve" sub-command
    serve_parser = sub_parsers.add_parser('serve', help='Serve the generated docs', parents=[base_parser])
    serve_parser.set_defaults(command='serve')
    serve_parser.add_argument('-p', '--port', type=int, default=8000, 
        help='Which port to run the server on')
    
    ## This adds the "new" sub-command
    new_parser = sub_parsers.add_parser('new', help='Generate starter site', parents=[base_parser])
    new_parser.set_defaults(command='new')
    new_parser.add_argument('-s', '--scaffold', type=str, default='blog-webawesome', 
        help='Which site scaffold to use to initialize your site')
    new_parser.add_argument('-y', '--yes', action='store_true', default=False, 
        help='Do not prompt for values, just use defaults')

    return root_parser.parse_args()


def main():
    args = get_args()

    if args.command == 'serve':
        DevServer(args).start()
        return
    
    if args.command == 'new':
        generate_scaffold(args)
        return
    
    run_pipeline(args)




from __future__ import annotations
from typing import TYPE_CHECKING, Dict
from os.path import isfile, join, abspath
from datetime import datetime
import logging
from git import Repo, InvalidGitRepositoryError
from git import BadName
from syrinx.exceptions import UnknownBranchError
if TYPE_CHECKING:
    from syrinx.config import BuildMetaInfo
    from syrinx.node import ContentNode
logger = logging.getLogger(__name__)


def read_branches(root_dir: str) -> Branches:
    """Read branch information from branches.toml file.
    
    Reads a TOML-formatted file containing branch names and their associated
    datetime values. Each line should be in the format: branch_name = datetime_string
    
    Args:
        root_dir: The root directory where branches.toml is located
        
    Returns:
        A Branches object containing a dictionary mapping branch names to datetime objects
    """
    cfg_fpath = join(abspath(root_dir), 'branches.toml')
    branches = dict()
    if isfile(cfg_fpath):
        with open(cfg_fpath) as fhandle:
            content = fhandle.read()
            for line in content.splitlines():
                if '=' not in line:
                    continue
                branch, dt_str, *_ = [p.strip() for p in line.split('=')]
                # Convert the datetime string to datetime object
                branches[branch] = datetime.fromisoformat(dt_str)
    if branches:
        logger.info(f'Found branches file with {len(branches)} branch')
    return Branches(branches)


def write_branches(branches: Dict[str, datetime], root_dir: str) -> None:
    """Write branch information to branches.toml file.
    
    Writes branch names and their associated datetime values to a TOML-formatted file.
    Each branch is written on a separate line in the format: branch_name = datetime_string
    Datetime objects are serialized using ISO 8601 format.
    
    Args:
        branches: Dictionary mapping branch names to datetime objects
        root_dir: The root directory where branches.toml will be created/overwritten
    """
    cfg_fpath = join(abspath(root_dir), 'branches.toml')
    with open(cfg_fpath, 'w') as fhandle:
        for branch, dt in branches.items():
            # Convert datetime to ISO format string
            dt_str = dt.isoformat()
            fhandle.write(f"{branch} = {dt_str}\n")


class Branches:
    
    def __init__(self, inner: Dict[str, datetime]) -> None:
        self.inner = inner
        self.active = None
        self.changed_files = set()

    def get_lastmodified(self, branch_name: str) -> datetime:
        """Get the last modified datetime for a specific branch.
        
        Args:
            branch_name: The name of the branch to look up
            
        Returns:
            The datetime object associated with the branch
            
        Raises:
            UnknownBranchError: If the branch is not found in self.inner
        """
        if branch_name not in self.inner:
            raise UnknownBranchError(
                f"Branch '{branch_name}' not found in branches data. "
                f"This may indicate that the branches.toml file is missing "
                f"or the branch entry is missing from the file."
            )
        return self.inner[branch_name]

    def update(self, meta: BuildMetaInfo, root_dir: str) -> None:
        """Update branch information with current build metadata.
        
        Uses git to determine the current branch name and records the build timestamp
        for that branch. If no branch is found, does nothing.
        Updates the internal branch dictionary and persists changes to branches.toml.
        Also caches the list of changed files for use in warnIfModifiedNodeHasOutdatedBranch.
        
        Args:
            meta: BuildMetaInfo object containing timestamp and other build metadata
            root_dir: The root directory where branches.toml will be written
        """
        # Convert root_dir to absolute path
        root_dir = abspath(root_dir)
        
        try:
            # Try to get the current branch name
            repo = Repo(root_dir, search_parent_directories=True)
            if repo.head.is_detached:
                # Detached HEAD state, no branch name available
                return
            branch_name = repo.active_branch.name
            
            # Cache changed files (both staged and unstaged, plus untracked)
            # Convert all paths to absolute paths
            try:
                repo_root = abspath(repo.working_dir)
                self.changed_files = set(join(repo_root, f) for f in repo.untracked_files)
                self.changed_files.update(join(repo_root, item.a_path) for item in repo.index.diff(None) if item.a_path)
                self.changed_files.update(join(repo_root, item.a_path) for item in repo.index.diff('HEAD') if item.a_path)
            except (AttributeError, TypeError):
                # Unable to get changed files (e.g., in tests with mocked repo)
                self.changed_files = set()
            
        except (InvalidGitRepositoryError, ValueError, TypeError, BadName):
            # Not a git repository or unable to determine branch
            return
        self.active = branch_name
        
        # Update the branch with the current build timestamp
        self.inner[branch_name] = meta.timestamp

        logger.info(f'Current branch: {branch_name}. Updating branches file.')
        
        # Write the updated branches to file if we have any entries
        if self.inner:
            write_branches(self.inner, root_dir)

    def warnIfModifiedNodeHasOutdatedBranch(self, node: ContentNode, branch: str):
        """Check if a modified node's file has uncommitted changes on a different branch.
        
        If the node's file has uncommitted changes and the specified branch doesn't
        match the currently active branch, logs a warning.
        Uses cached changed_files populated by update() for efficiency.
        
        Args:
            node: The ContentNode to check
            branch: The branch name specified in the node's LastModifiedBranch
        """
        # Only check if we have an active branch to compare against
        if self.active is None:
            return
        
        # If the branch matches the active branch, no warning needed
        if branch == self.active:
            return
        
        source_path_clean = node.source_path.lstrip('/')
        file_path = abspath(join(node.meta.root_dir, 'content', source_path_clean))
        
        # Check if the file is in the cached changed files
        if file_path in self.changed_files:
            logger.warning(
                f"File '{file_path}' has uncommitted changes "
                f"but references LastModifiedBranch '{branch}' while currently on branch '{self.active}'"
            )



from __future__ import annotations
import logging
from typing import Optional, TYPE_CHECKING
from os.path import isfile, join, abspath
from datetime import datetime, UTC
from importlib.metadata import version
from syrinx.branches import read_branches
if TYPE_CHECKING:
    from argparse import Namespace
    from syrinx.branches import Branches


class BuildMetaInfo:

    def __init__(self, config: SyrinxConfiguration, root_dir: str) -> None:
        self.environment = config.environment
        self.timestamp = datetime.now(tz=UTC)
        self.syrinx_version = version('syrinx')
        self.root_dir = root_dir



class SyrinxConfiguration:
    clean: bool
    domain: Optional[str]
    environment: str
    leaf_pages: bool
    sitemap: str
    urlformat: str
    verbose: bool
    branches: Branches
    meta: BuildMetaInfo

    def __str__(self) -> str:
        lines = []
        KEYS = ('clean', 'domain', 'environment', 'leaf_pages', 
                'sitemap', 'urlformat', 'verbose')
        for key in KEYS:
            val = getattr(self, key)
            if isinstance(val, str):
                val = f'"{val}"'
            if isinstance(val, bool):
                val = str(val).lower()
            lines.append(f'\t{key} = {val}')
        return '\n'.join(lines)


def configure(args: Namespace) -> SyrinxConfiguration:
    config = SyrinxConfiguration()
    config.clean = True
    config.domain = None
    config.environment = 'default'
    config.leaf_pages = False
    config.sitemap = 'opt-out'
    config.urlformat = 'filesystem'
    config.verbose = False



    root_dir = getattr(args, 'dir', '.')
    cfg_fpath = join(abspath(root_dir), 'syrinx.cfg')
    if isfile(cfg_fpath):
        with open(cfg_fpath) as fhandle:
            content = fhandle.read()
            #print(content)
            for line in content.splitlines():
                if '=' not in line:
                    continue
                key, val, *_ = [p.strip() for p in line.split('=')]
                val = val.strip('"')
                if key == 'clean':
                    config.clean = val.lower() == 'true'
                elif key == 'domain':
                    config.domain = val
                elif key == 'environment':
                    config.environment = val
                elif key == 'leaf_pages':
                    config.leaf_pages = val.lower() == 'true'
                elif key == 'sitemap':
                    if val not in ('opt-in', 'opt-out'):
                        raise ValueError('Configuration option "sitemap" must'
                                         ' be one of "opt-in", "opt-out".')
                    config.sitemap = val
                elif key == 'urlformat':
                    config.urlformat = val
                elif key == 'verbose':
                    config.verbose = val.lower() == 'true'
                else:
                    raise ValueError(f'Unknown configuration entry: {key}')

    for key in ('clean', 'domain', 'verbose', 'environment', 'leaf_pages', 'urlformat'):
        if hasattr(args, key):
            setattr(config, key, getattr(args, key))

    if config.verbose:
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger(__name__)
        logger.info('Configuration:\n'+str(config))

    config.meta = BuildMetaInfo(config, root_dir)
    config.branches = read_branches(root_dir)
    config.branches.update(config.meta, root_dir)
    return config



"""
    The user calls `syrinx new` to generate a starting workspace for their static site.

    Collect values for variables from the user with the "rich" library.
    Each should have a default.

    Available scaffolds are in the scaffolds directory. Currently there is just one 
    but I will add more later.

    Files with placeholders ( marked with double brackets ) should be processed with jinja2
    by passing in a "scaffold" variable with the template variables as properties.

    Files without placeholders could simply be copied,
    but for simplicity it may make sense to just process them all with jinja2.

    a) which scaffold (or pass in as arg..)
    c) site title
    d) site description
    e) config options for syrinx.cfg
    f) template specific vars (to be defined in a json file in the scaffold directory):
        1) web awesome pro kit code "fa-kit-code"

    display generated files with as tree 
    give suggestion for next cli command: syrinx serve <dir>
"""
from __future__ import annotations
from typing import TYPE_CHECKING
import os
import json
from pathlib import Path
from os.path import abspath, join, dirname
from importlib.metadata import version
from jinja2 import Environment, Template
from rich.console import Console
from rich.prompt import Prompt
from rich.tree import Tree
if TYPE_CHECKING:
    from argparse import Namespace


## Common variables needed for all scaffolds
COMMON_VARIABLES = [
    {
        "name": "title",
        "prompt": "Site title",
        "default": "My Site"
    },
    {
        "name": "description",
        "prompt": "Site description",
        "default": "A site built with Syrinx"
    },
    {
        "name": "domain",
        "prompt": "Domain name",
        "default": "example.com"
    },
    {
        "name": "syrinx_version",
        "prompt": "Syrinx version",
        "default": version('syrinx')
    }
]


def generate_scaffold(args: Namespace):
    """Generate a new site from a scaffold template."""
    
    console = Console()
    
    ## Target directory (where to output files)
    target_dir = abspath(args.dir)
    
    ## Scaffold choice
    scaffold_name = args.scaffold
    
    ## Get path to scaffolds directory
    syrinx_root = dirname(dirname(abspath(__file__)))
    scaffolds_dir = join(syrinx_root, 'scaffolds')
    scaffold_dir = join(scaffolds_dir, scaffold_name)
    
    ## Check if scaffold exists
    if not os.path.exists(scaffold_dir):
        console.print(f"[red]Error: Scaffold '{scaffold_name}' not found at {scaffold_dir}[/red]")
        return
    
    ## Load scaffold metadata
    scaffold_json_path = join(scaffold_dir, 'scaffold.json')
    if not os.path.exists(scaffold_json_path):
        console.print(f"[red]Error: scaffold.json not found in {scaffold_dir}[/red]")
        return
    
    with open(scaffold_json_path, 'r') as f:
        scaffold_meta = json.load(f)
    
    ## Display scaffold info
    console.print(f"\n[bold cyan]Creating new site: {scaffold_meta['name']}[/bold cyan]")
    console.print(f"[dim]{scaffold_meta['description']}[/dim]\n")
    
    ## Merge common variables with scaffold-specific variables
    all_variables = COMMON_VARIABLES + scaffold_meta.get('variables', [])
    
    ## Collect variable values from user
    scaffold_vars = {}
    
    if args.yes:
        ## Use defaults without prompting
        console.print("[yellow]Using default values (--yes flag)[/yellow]\n")
        for var in all_variables:
            scaffold_vars[var['name']] = var['default']
            console.print(f"  {var['prompt']}: [cyan]{var['default']}[/cyan]")
    else:
        ## Prompt user for each variable
        console.print("[bold]Please provide the following information:[/bold]\n")
        for var in all_variables:
            value = Prompt.ask(
                f"  {var['prompt']}", 
                default=var['default']
            )
            scaffold_vars[var['name']] = value
    
    console.print()
    
    ## Check if target directory exists
    if os.path.exists(target_dir) and os.listdir(target_dir):
        console.print(f"[yellow]Warning: Directory {target_dir} already exists and is not empty[/yellow]")
        if not args.yes:
            proceed = Prompt.ask("  Continue anyway?", choices=["y", "n"], default="n")
            if proceed.lower() != 'y':
                console.print("[red]Cancelled[/red]")
                return
    
    ## Create target directory if it doesn't exist
    os.makedirs(target_dir, exist_ok=True)
    
    ## Use alternative syntax with brackets for scaffolding phase
    env = Environment(
        block_start_string='[%',
        block_end_string='%]',
        variable_start_string='[[',
        variable_end_string=']]',
        comment_start_string='[#',
        comment_end_string='#]'
    )
    
    ## Process all files from scaffold
    console.print("[bold green]Generating files...[/bold green]\n")
    generated_files = []
    
    for root, dirs, files in os.walk(scaffold_dir):
        ## Skip scaffold.json
        if 'scaffold.json' in files:
            files.remove('scaffold.json')
        
        ## Calculate relative path from scaffold_dir
        rel_dir = os.path.relpath(root, scaffold_dir)
        if rel_dir == '.':
            target_subdir = target_dir
        else:
            target_subdir = join(target_dir, rel_dir)
        
        ## Create target subdirectory
        os.makedirs(target_subdir, exist_ok=True)
        
        ## Process each file
        for filename in files:
            source_file = join(root, filename)
            target_file = join(target_subdir, filename)
            
            ## Read source file content
            try:
                with open(source_file, 'r', encoding='utf-8') as f:
                    content = f.read()
                
                ## Process with jinja2
                template = env.from_string(content)
                processed_content = template.render(scaffold=scaffold_vars)
                
                ## Write to target file
                with open(target_file, 'w', encoding='utf-8') as f:
                    f.write(processed_content)
                
                ## Track generated file
                rel_path = os.path.relpath(target_file, target_dir)
                generated_files.append(rel_path)
                
            except UnicodeDecodeError:
                ## Binary file - just copy it
                import shutil
                shutil.copy2(source_file, target_file)
                rel_path = os.path.relpath(target_file, target_dir)
                generated_files.append(rel_path)
            except Exception as e:
                console.print(f"[red]Error processing {filename}: {e}[/red]")
    
    ## Display generated files as tree
    console.print("[bold green]✓ Site generated successfully![/bold green]\n")
    console.print("[bold]Generated files:[/bold]\n")
    
    tree = Tree(f"[bold blue]{os.path.basename(target_dir)}[/bold blue]")
    
    ## Build tree structure
    tree_nodes = {'.': tree}
    for file_path in sorted(generated_files):
        parts = Path(file_path).parts
        
        if not parts:
            continue
            
        current_path = '.'
        
        ## Create intermediate directory nodes
        for i, part in enumerate(parts[:-1]):
            parent_path = current_path
            current_path = str(Path(current_path) / part)
            
            if current_path not in tree_nodes:
                tree_nodes[current_path] = tree_nodes[parent_path].add(f"[blue]{part}/[/blue]")
        
        ## Add file node
        if len(parts) > 0:
            parent_path = str(Path(*parts[:-1])) if len(parts) > 1 else '.'
            tree_nodes[parent_path].add(f"[green]{parts[-1]}[/green]")
    
    console.print(tree)
    console.print()
    
    ## Display next steps
    console.print("[bold cyan]Next steps:[/bold cyan]")
    console.print(f"  1. cd {target_dir}")
    console.print(f"  2. syrinx serve")
    console.print()



from __future__ import annotations
from typing import TYPE_CHECKING, List, Dict, Optional
from sys import maxsize as SYS_MAX_SIZE
from os.path import dirname, basename
from datetime import datetime
if TYPE_CHECKING:
    from syrinx.config import SyrinxConfiguration, BuildMetaInfo


def makeBranchNode(config: SyrinxConfiguration, name: str) -> ContentNode:
    node = ContentNode(config)
    node.name = name
    node.isLeaf = False
    return node


def makeLeafNode(config: SyrinxConfiguration) -> ContentNode:
    node = ContentNode(config)
    node.isLeaf = True
    return node


class ContentNode:
    name: str
    leaves: List[ContentNode]
    branches: List[ContentNode]
    content_html: str
    front: Dict[str, str]
    path: str
    source_path: str
    config: SyrinxConfiguration
    isLeaf: bool
    fpath: str

    def __init__(self, config: SyrinxConfiguration):
        self.leaves = []
        self.branches = []
        self.front = {}
        self.config = config
        self.isLeaf = False
        self.source_path = ''

    def setContent(self, fpath: str, front: Dict[str, str], html: str) -> None:
        self.fpath = fpath
        self.front = front
        self.content_html = html
        self.source_path = fpath
        self.path = dirname(fpath)
        if self.path == '/':
            self.path = ''
        if self.isLeaf:
            fparts = basename(fpath).split('.')
            self.name = fparts[0]

    @property
    def meta(self) -> BuildMetaInfo:
        return self.config.meta

    @property
    def sequenceNumber(self) -> int:
        if 'SequenceNumber' in self.front:
            return self.front['SequenceNumber']
        else:
            return SYS_MAX_SIZE
        
    @property
    def buildPage(self) -> bool:
        if self.isLeaf and not self.config.leaf_pages:
            return False
        elif not self.source_path:
            return False
        else:
            return True

    @property
    def title(self) -> str:
        if 'Title' in self.front:
            return self.front['Title']
        else:
            return self.name.replace('_', ' ').title()
        
    @property
    def address(self) -> Optional[str]:
        """Full, canonical URL of this node

        Returns None if no domain configured. Applies one of three styles
        based on the *urlformat* config directive.
        """
        if self.config.domain is None:
            ## can't make an address if we don't know the domain
            return
        is_directory = any(self.branches) or any(self.leaves)
        trail = ''
        if len(self.name):
            if self.isLeaf:
                trail = f'/{self.name}'
                if self.config.urlformat == 'filesystem':
                    trail += '.html'
                elif self.config.urlformat == 'mkdocs':
                    trail += '/'
            elif is_directory or self.config.urlformat != 'clean':
                trail = '/'
        return f'https://{self.config.domain}{self.path}{trail}'

    @property
    def lastModified(self) -> Optional[datetime]:
        # First check for direct LastModified entry
        if 'LastModified' in self.front:
            # Convert string datetime to datetime object
            return datetime.fromisoformat(self.front['LastModified'])
        
        # If not present, check for LastModifiedBranch entry
        if 'LastModifiedBranch' in self.front:
            branch_name = self.front['LastModifiedBranch']
            # warn if modified, but on new branch
            self.config.branches.warnIfModifiedNodeHasOutdatedBranch(self, branch_name)
            # Return datetime object directly from branches
            return self.config.branches.get_lastmodified(branch_name)

        return None
    
    @property
    def includeInSitemap(self) -> bool:
        if not self.buildPage:
            return False
        if self.address is None:
            return False
        default = {'opt-in': False, 'opt-out': True}[self.config.sitemap]
        iis = self.front.get('IncludeInSitemap', default)
        if not isinstance(iis, bool):
            raise ValueError('IncludeInSitemap config entry must be boolean')
        return iis



from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Tuple
from os.path import dirname, basename, join
from os import walk
from tomllib import loads as read_toml
from yaml import safe_load as read_yaml
import logging
from markdown import markdown
from syrinx.exceptions import ContentError
from syrinx.node import ContentNode, makeBranchNode, makeLeafNode
if TYPE_CHECKING:
    from syrinx.config import SyrinxConfiguration
logger = logging.getLogger(__name__)
"""
This section is just about reading and interpreting the content
"""


def reorder_children(node: ContentNode):
    node.leaves = sorted(node.leaves, key=lambda n: (n.sequenceNumber, n.name))
    node.branches = sorted(node.branches, key=lambda n: (n.sequenceNumber, n.name))
    for child in node.branches:
        reorder_children(child)


def read_file(fpath: str) -> Tuple[Dict, str]:
    """Read a single markdown content file and 
    return frontmatter as dictionary and contents as string.

    Can read either toml or yaml-style frontmatter

    Args:
        fpath (str): Full path to markdown file

    Returns:
        Tuple[Dict, str]: Dictionary with frontmatter attributes, and 
            string with main contents
    """
    with open(fpath) as fhandle:
        lines = fhandle.readlines()
    formats = [('+++', read_toml), ('---', read_yaml)]
    for marker, parser in formats:
        marker_lines = [l for (l, line) in enumerate(lines) if line.strip() == marker]
        if len(marker_lines) >= 2:
            fm_string = ''.join(lines[1:marker_lines[1]])
            fm_dict = parser(fm_string)
            md_content = ''.join(lines[marker_lines[1]+1:])
            break
    else:
        fm_dict = dict()
        md_content = ''.join(lines)
        logger.warning(f'No frontmatter found in {fpath}')
    return fm_dict, md_content


def read(root_dir: str, config: SyrinxConfiguration) -> ContentNode:

    content_dir = join(root_dir, 'content')

    tree: Dict[str, ContentNode] = dict()
    root = makeBranchNode(config, '')
    for (dirpath, _, fnames) in walk(content_dir):

        indexNode = makeBranchNode(config, basename(dirpath))
        if dirpath == content_dir:
            indexNode = root
            if 'index.md' not in fnames:
                raise ContentError('root index file missing') 
        else:
            parent = tree[dirname(dirpath)]
            parent.branches.append(indexNode)
        tree[dirpath] = indexNode

        ## ideally process the index page first (not sure if this is necessary?)
        if 'index.md' in fnames:
            fnames.insert(0, fnames.pop(fnames.index('index.md')))

        for fname in fnames:
            if not fname.endswith('.md'):
                continue

            fpath = join(dirpath, fname)
            fm_dict, md_content = read_file(fpath)

            if fname == 'index.md':
                node = indexNode
            else:
                node = makeLeafNode(config)
                indexNode.leaves.append(node)

            node.setContent(fpath.replace(content_dir, ''), fm_dict, markdown(md_content, extensions=['fenced_code', 'tables']))
            logger.info(f'Read {node.source_path}')

    reorder_children(root)

    return root



"""Convert data files to content

The archetype used will be based on the filename of the data file.
Each row is considered one record.
The header (1st) row will be used as keys.
The first column is used as name for the content item.
Each column will be converted to a variable in the front matter 
"""
from __future__ import annotations
from typing import TYPE_CHECKING, Dict
from os.path import join, isdir, basename
from os import makedirs, remove
import logging
from glob import glob
from jinja2 import Environment, FileSystemLoader, select_autoescape
from pandas import read_csv
from numpy import nan
if TYPE_CHECKING:
    from jinja2 import Template
    from syrinx.config import SyrinxConfiguration
logger = logging.getLogger(__name__)

def preprocess(root_dir: str, config: SyrinxConfiguration) -> None:

    assert isdir(root_dir)

    archetype_files = glob(join(root_dir, 'archetypes', '*.md'))
    archetypes: Dict[str, Template] = dict()
    env = Environment(
        loader=FileSystemLoader(join(root_dir, 'archetypes')),
        autoescape=select_autoescape()
    )

    for fpath in archetype_files:
        fname = basename(fpath)
        archetype_name = fname.split('.')[0]
        archetypes[archetype_name] = env.get_template(fname)


    data_files = glob(join(root_dir, 'data', '*.tsv'))

    for fpath in data_files:
        archetype_name = basename(fpath).split('.')[0]
        if archetype_name not in archetypes:
            raise ValueError(f'No archetype for {archetype_name}')
        archetype = archetypes[archetype_name]

        collection_dir = join(root_dir, 'content', archetype_name)
        makedirs(collection_dir, exist_ok=True)

        if config.clean:
            stale_content = glob("*.md", root_dir=collection_dir)
            [remove(join(collection_dir, filename)) for filename in stale_content if "index" not in filename]

        df = read_csv(fpath, sep='\t', index_col=0)

        ## get rid of whitespace in columns:
        df.columns = [col.strip().replace(' ', '_') for col in df.columns]

        ## turns nans to None
        df.replace([nan], [None], inplace=True)

        ## add sequence number if not provided
        if 'SequenceNumber' not in df.columns:
            df['SequenceNumber'] = list(range(len(df)))

        df['Archetype'] = [archetype_name]*len(df)

        logger.info(f'Preprocessing {archetype_name}')

        for label, row in df.iterrows():
            output = archetype.render(dict(row) | {df.index.name: label})
            with open(join(collection_dir, f'{label}.md'), 'w') as fhandle:
                fhandle.write(output)



from __future__ import annotations
from typing import TYPE_CHECKING, Callable
from os.path import isdir, join, isfile
import shutil, os, logging
from jinja2 import Environment, FileSystemLoader, select_autoescape
from syrinx.exceptions import ThemeError
from syrinx.sitemap import collect_urls, generate_sitemap
if TYPE_CHECKING:
    from syrinx.read import ContentNode
    from syrinx.config import SyrinxConfiguration
logger = logging.getLogger(__name__)

def choose_template_file(
            node: ContentNode,
            isfile: Callable[[str], bool],
            dir: str
        ) -> str:
    name = node.name or 'root'
    template_names = (name, 'leaf', 'page') if node.isLeaf else (name, 'page')
    for tem_name in template_names:
        if isfile(join(dir, f'{tem_name}.jinja2')):
            return f'{tem_name}.jinja2'
    else:
        raise ThemeError(f'Missing template for "{node.name}"')


def dir_exists_not_empty(path: str) -> bool:
    if isdir(path):
        if len(os.listdir(path)):
            return True
    return False


def build_node(
        node: ContentNode,
        root: ContentNode,
        parent_path: str,
        template_dir: str,
        env: Environment
    ):
    """Recursive function to render page, then move on to children
    """
    node_path = join(parent_path, node.name)
    os.makedirs(node_path, exist_ok=True)
    if node.buildPage:
        fname_tem = choose_template_file(node, isfile, template_dir)
        page_template = env.get_template(fname_tem)
        html = page_template.render(index=node, root=root)
        out_fpath = join(node_path, f'{node.name}.html' if node.isLeaf else 'index.html')
        with open(out_fpath, 'w') as fhandle:
            fhandle.write(html)
        rel_path = out_fpath.replace(out_fpath.split('dist/')[0], '')
        logger.info(f'Built {rel_path}')

    for child in node.branches+node.leaves:
        build_node(child, root, node_path, template_dir, env)


def build(root: ContentNode, root_dir: str, config: SyrinxConfiguration):

    assert isdir(root_dir)
            
    theme_dir = join(root_dir, 'theme')
    template_dir = join(theme_dir, 'templates')
    env = Environment(
        loader=FileSystemLoader(template_dir),
        autoescape=select_autoescape()
    )

    ## locate and clear target directory
    dist_dir = join(root_dir, 'dist')
    if isdir(dist_dir):
        shutil.rmtree(dist_dir)
    os.makedirs(dist_dir, exist_ok=True)

    build_node(root, root, dist_dir, template_dir, env)

    dist_assets_dir = join(dist_dir, 'assets')

    ## copy theme assets tree to dist 
    theme_assets_dir = join(theme_dir, 'assets')
    if dir_exists_not_empty(theme_assets_dir):
        shutil.copytree(theme_assets_dir, dist_assets_dir)

    ## copy assets tree to dist 
    content_assets_dir = join(root_dir, 'assets')
    if dir_exists_not_empty(content_assets_dir):
        shutil.copytree(content_assets_dir, dist_assets_dir, dirs_exist_ok=True)

    sitemap_fpath = join(dist_dir, 'sitemap.xml')
    urls = collect_urls(root)
    if urls:
        with open(sitemap_fpath, 'w') as fhandle:
            fhandle.write(generate_sitemap(urls))
        logger.info(f'Created sitemap.xml')



from argparse import Namespace
from os.path import abspath, isdir
from syrinx.build import build
from syrinx.read import read
from syrinx.preprocess import preprocess
from syrinx.config import configure


def run_pipeline(args: Namespace):
    """Execute the full build pipeline.
    
    Args:
        root_dir: Root directory to build from
        args: Command-line arguments
    
    Returns:
        The root ContentNode
    """
    root_dir = abspath(args.dir)
    assert isdir(root_dir)
    config = configure(args)
    preprocess(root_dir, config)
    root = read(root_dir, config)
    build(root, root_dir, config)
    return root
